"""
EvenVizion library.
https://github.com/AIHunters/EvenVizion

Supporting paper at:
https://github.com/AIHunters/EvenVizion/blob/master/EvenVizion-video_based_camera_localization_component.pdf
This is licensed under an MIT license. See the LICENSE.md file
for more information.

To describe a frame with the help of key points of different types.
"""
import imutils
import cv2
import numpy as np

from evenvizion.processing.matching import KeyPoints, NoMatchesException
from evenvizion.processing.utils import remove_double_matching


class FrameProcessing:
    """
    A class used to describe a frame with the help of key points of different types.

    Attributes
    ----------
    frame: np.array
        An array of n special points from the first frame.

    Methods
    -------
    detect_and_describe_features(frame)
        To find key points on a frame and return their coordinates and descriptors.
    concatenate_all_features_types(frame)
        To concatenate all key points types in one array.
    """

    def __init__(self, frame, features_type_list=None):
        self.isv3 = imutils.is_cv3(or_better=True)
        self.frame = frame
        self.features_types = features_type_list or ["SURF", "SIFT", "ORB"]

    def detect_and_describe_features(self, features_name):
        """ To find key points on a frame and return their coordinates and descriptors.

        Parameters
        ----------
        features_name: str
            The name of features type.
                example: "ORB"

        Returns
        ----------
        coordinates: np.array
            An array of x and y coordinates of each key point.

        descriptors: np.array
            An array of descriptors of each key point.
        """
        if features_name == "ORB":
            orb = cv2.ORB_create()
            (coordinates, descriptors) = orb.detectAndCompute(self.frame, None)
        elif features_name == "SIFT":
            sift = cv2.xfeatures2d.SIFT_create()
            (coordinates, descriptors) = sift.detectAndCompute(self.frame, None)
        elif features_name == "SURF":
            surf = cv2.xfeatures2d.SURF_create(extended=1, hessianThreshold=400)
            (coordinates, descriptors) = surf.detectAndCompute(self.frame, None)
        else:
            raise ValueError("You need to choose descriptors type")
        coordinates = np.float32([kp.pt for kp in coordinates])
        return coordinates, descriptors

    def concatenate_all_features_types(self, acceding_image):
        """ To concatenate all key points types in one array.

        Parameters
        ----------
        acceding_image: np.array
            The frame to concatenate.

        Returns
        ----------
        all_matching_pts_a: np.array
            An array of x and y coordinates of key points on static objects from self.frame.

        all_matching_pts_b: np.array
            An array of x and y coordinates of key points on static objects from an acceding image.
        """
        all_static_pts_a_with_repetitions = []
        all_static_pts_b_with_repetitions = []
        for feature_type in self.features_types:
            (coords_a, descriptors_a) = self.detect_and_describe_features(feature_type)
            (coords_b, descriptors_b) = acceding_image.detect_and_describe_features(feature_type)
            kps_a = KeyPoints(coords_a, descriptors_a)
            kps_b = KeyPoints(coords_b, descriptors_b)
            static_pts_a, static_pts_b = kps_a.match_static_kps(kps_b)
            all_static_pts_a_with_repetitions.extend(static_pts_a)
            all_static_pts_b_with_repetitions.extend(static_pts_b)
        if all_static_pts_a_with_repetitions is None or all_static_pts_b_with_repetitions is None:
            raise NoMatchesException("can't find keypoints that lie on static objects ",
                                     "couldn't process")
        all_matching_pts_a, all_matching_pts_b = \
            remove_double_matching(all_static_pts_a_with_repetitions,
                                   all_static_pts_b_with_repetitions)
        if all_static_pts_a_with_repetitions is None or all_static_pts_b_with_repetitions is None:
            raise NoMatchesException("can't find keypoints that lie on static objects ",
                                     "couldn't process")
        return all_matching_pts_a, all_matching_pts_b



"""
EvenVizion library.
https://github.com/AIHunters/EvenVizion

Supporting paper at:
https://github.com/AIHunters/EvenVizion/blob/master/EvenVizion%20-%20video%20based%20camera%20localization%20component.pdf

This is licensed under an MIT license. See the LICENSE.md file
for more information.

This is an example of the comparison of visualizations between the original video and stabilized video.
"""

__version__ = "0.9"

__all__ = ['processing_visualization', 'stabilization']

from .processing_visualization import *
from .stabilization import *



"""
EvenVizion library.
https://github.com/RnD-Oxagile/EvenVizion

Supporting paper at:
https://github.com/AIHunters/EvenVizion/blob/master/EvenVizion-video_based_camera_localization_component.pdf

This is licensed under an MIT license. See the README.md file
for more information.

This is an example of getting homography matrix for each frame, getting heatmap visualization.
If you specify --path_to_original_coordinate,
original coordinates will be recalculated to a fixed coordinate system.
"""

import argparse
import json
import os
import sys

import cv2
import numpy as np

main_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(main_dir)

main_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.append(main_dir)

from evenvizion.processing.fixed_coordinate_system import from_original_to_fix
from evenvizion.processing.utils import read_homography_dict, superposition_dict, \
    are_infinity_coordinates, read_json_with_coordinates
from evenvizion.processing.video_processing import get_homography_dict
from evenvizion.visualization.processing_visualization import \
    heatmap_video_processing, comparison_original_with_fixed_coordinate_video_processing



def visualize_heatmap(save_folder, path_to_video, path_to_homography_dict):
    """ To create heatmap visualization.

    Parameters
    ----------
    save_folder: str
        The path to save pictures with heatmap visualization.

    path_to_video: str
        The path to the analyzed video.

    path_to_homography_dict: str
        The path to the json file with homography matrix,
        which describes movement between consistent frames on video.

    """
    cap = cv2.VideoCapture(path_to_video)
    homography_matrices, _ = read_homography_dict(path_to_homography_dict)
    reformat_homography_dict = superposition_dict(homography_matrices)
    heatmap_save_folder = "{}/heatmap_visualization".format(save_folder)
    if not os.path.exists(heatmap_save_folder):
        os.mkdir(heatmap_save_folder)
    max_movement = heatmap_video_processing(
        reformat_homography_dict, cap, heatmap_save_folder)
    with open("{}/{}.txt".format(args.save_folder, "metrics_file"), "w") as txt_:
        txt_.write("Maximum movement during the entire video: {}".format(np.max(max_movement)))
        if are_infinity_coordinates(max_movement):
            txt_.write("There are some frames with undefined coordinates")


def get_fixed_coordinate(path_to_video, path_to_homography_dict, original_shape):
    """ To recalculate and visualize original coordinates to fixed.

    Parameters
    ----------
    path_to_video: str
        The path to the analyzed video.

    path_to_homography_dict: str
        The path to the json file with homography matrix,
        which describes movement between consistent frames on video.

    original_shape: list, or np.array
        An original frame shape, at the first place - height, the second - width.
        example:  [224,400]

    """
    dict_with_homography_matrix, resize_info = read_homography_dict(path_to_homography_dict)
    reformat_homography_dict = superposition_dict(dict_with_homography_matrix)
    original_coordinates = read_json_with_coordinates(args.path_to_original_coordinate)
    recalculated_coordinates = from_original_to_fix(original_coordinates,
                                                    reformat_homography_dict, original_shape,
                                                    [resize_info["h"], resize_info["w"]])
    cap = cv2.VideoCapture(path_to_video)
    comparison_save_folder = "{}/fixed_coordinate_system_visualization".format(args.save_folder)
    comparison_original_with_fixed_coordinate_video_processing(cap, original_coordinates,
                                                               recalculated_coordinates,
                                                               comparison_save_folder,
                                                               resize_info)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='custom arguments')
    parser.add_argument('--path_to_video', type=str, default="test_video/test_video.mp4")
    parser.add_argument('--experiment_name', type=str, default="test_video_processing")
    parser.add_argument('--resize_width', type=int, help="wights to resize image", default=400)
    parser.add_argument('--path_to_original_coordinate',
                        help="path to json with original coordinate",
                        default="test_video/original_coordinates.json")
    parser.add_argument('--none_H_processing', help="If True we use H_prev as H, False- do nothing",
                        default=True)
    parser.add_argument('--heatmap_visualization',
                        help="Getting heatmap visualization",
                        default=True)
    parser.add_argument('--show_matching_visualization',
                        help="Getting matching visualization",
                        default=True)

    args = parser.parse_args()
    args.experiment_folder = os.getcwd()
    args.save_folder = args.experiment_folder + "/" + args.experiment_name
    if not os.path.exists(args.save_folder):
        os.makedirs(args.save_folder)

    args.save_folder = args.save_folder + '/{}'.format(os.path.split(args.path_to_video)[-1].split(".")[0])
    if not os.path.exists(args.save_folder):
        os.makedirs(args.save_folder)
    path_to_save_matching = None
    if args.show_matching_visualization:
        path_to_save_matching = os.path.join(args.save_folder, "matching_visualization")
        if not os.path.exists(path_to_save_matching):
            os.makedirs(path_to_save_matching)

    cap = cv2.VideoCapture(args.path_to_video)
    original_img_w = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    original_img_h = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    original_shape = [original_img_h, original_img_w]
    result = get_homography_dict(cap, matching_path=path_to_save_matching,
                                 none_H_processing=args.none_H_processing)

    with open("{}/dict_with_homography_matrix.json".format(args.save_folder), "w") as json_:
        json.dump(result, json_)
    path_to_homography_dict = "{}/dict_with_homography_matrix.json".format(args.save_folder)

    if args.heatmap_visualization:
        visualize_heatmap(args.save_folder, args.path_to_video, path_to_homography_dict)
    if args.path_to_original_coordinate:
        get_fixed_coordinate(args.path_to_video, path_to_homography_dict, original_shape)



"""
EvenVizion library.
https://github.com/AIHunters/EvenVizion

Supporting paper at:
https://github.com/AIHunters/EvenVizion/blob/master/EvenVizion-video_based_camera_localization_component.pdf
This is licensed under an MIT license. See the LICENSE.md file
for more information.

To describe the key points from a frame.
"""

import cv2
import numpy as np

from evenvizion.processing.constants import THRESHOLD_FOR_FIND_HOMOGRAPHY, \
    LOWES_RATIO, MINIMUM_MATCHING_POINTS
from evenvizion.processing.utils import remove_double_matching, \
    find_point_displacement, get_largest_group_points


class NoMatchesException(Exception):
    """
        The class used to handle the exception.

        Attributes
        ----------
        reason: : str
            Exception reason.

        description: str
            Exception description.

        Methods
        ----------
    """

    def __init__(self, reason, description="no matches found"):
        self.reason = reason
        self.description = description
        super().__init__(description)

    def __str__(self):
        return f'{self.reason} -> {self.description}'


class KeyPoints:
    """
        A class used to describe key points.

        Attributes
        ----------
        coordinates: : np.array
            An array of key points coordinates.

        descriptors: : np.array
            An array of key points descriptors.

        Methods
        -------
        match_kps(acceding_kps: KeyPoints)
            To match key points using their descriptors.
            Find 2 nearest neighbours for every key point,
            and using Lowe's ratio test to delete false matchings.

        match_static_kps(acceding_kps: KeyPoints)
            To match key points which are on a static object using their descriptors.

    """

    def __init__(self, coordinates, descriptors):
        self.coordinates = coordinates
        self.descriptors = descriptors

    def match_kps(self, acceding_kps, ratio=LOWES_RATIO, min_matching_pts=MINIMUM_MATCHING_POINTS):
        """ To match key points using their descriptors.
         Find 2 nearest neighbours for every key point,
         and using Lowe's ratio test to delete false matchings.

        Parameters
        ----------
        acceding_kps: KeyPoints
            Key points should be matched with self key points.

        ratio: float, optional
            The ratio for Lowe's test.

        min_matching_pts: int, optional
            The minimum number of matching points you want to find.

        Returns
        ----------
        pts_a: np.array
            An array of x and y coordinates of each self key point
            which matches with acceding key points.

        pts_b: np.array
            An array of x and y coordinates of each acceding key point
            which matches with self key points.
        """

        matcher = cv2.DescriptorMatcher_create(
            "BruteForce")
        if self.descriptors is None:
            raise NoMatchesException("self.descriptors is None", "couldn't process")
        if acceding_kps.descriptors is None:
            raise NoMatchesException("kps.descriptors is None", "couldn't process")
        raw_matches = matcher.knnMatch(self.descriptors, acceding_kps.descriptors, 2)
        if raw_matches is None:
            raise NoMatchesException("raw_matches is None",
                                     "can't find matches points between this frames")
        matches = lowes_ratio_test(raw_matches, ratio)
        if len(matches) < min_matching_pts:
            raise NoMatchesException("len(matches) {} < "
                                     "min_matching_pts {}".format(len(matches), min_matching_pts),
                                     "couldn't process")
        pts_a = np.float32([self.coordinates[i] for (_, i) in matches])
        pts_b = np.float32([acceding_kps.coordinates[i] for (i, _) in matches])
        pts_a, pts_b = remove_double_matching(pts_a, pts_b)
        if pts_a is None:
            raise NoMatchesException("problem with finding points at the first frame",
                                     "couldn't process")
        if pts_b is None:
            raise NoMatchesException("problem with finding points at the second frame",
                                     "couldn't process")
        if len(pts_a) != len(pts_b):
            raise NoMatchesException("lengths between matching points are different",
                                     "couldn't process")
        return pts_a, pts_b

    def match_static_kps(self, acceding_kps, reproj_thresh=THRESHOLD_FOR_FIND_HOMOGRAPHY):
        """ To match key points which are on static object using their descriptors.

        Parameters
        ----------
        acceding_kps: KeyPoints
            Key points should be matched with self key points.

        reproj_thresh: float, or int
            The threshold for OpenCV findHomography() function.

        Returns
        ----------
        static_pts_a: np.array
            An array of x and y coordinates of each self key point on a static object
            which can be matched with acceding key points.

        static_pts_b: np.array
            An array of x and y coordinates of each acceding key point on a static object
            which can be matched with self key points.
        """
        matching_keypoints = self.match_kps(acceding_kps)
        if matching_keypoints is None:
            raise NoMatchesException("can't find matching points", "couldn't process")
        pts_a, pts_b = matching_keypoints
        (matrix_H, _) = cv2.findHomography(np.array(pts_a), np.array(pts_b),
                                           cv2.RANSAC, reproj_thresh)
        if matrix_H is None:
            raise NoMatchesException("can't find homography matrix", "couldn't process")

        r_moving_dict = find_point_displacement(matrix_H, pts_a, pts_b)
        static_pts_a, static_pts_b = get_largest_group_points(r_moving_dict, pts_a, pts_b)
        return static_pts_a, static_pts_b


def lowes_ratio_test(raw_matches, ratio=LOWES_RATIO):
    """ To delete false matchings using Lowe's ratio test.

    Parameters
    ----------
    raw_matches: list
        A list of key points with 2 nearest (similar descriptors) neighbours.

    ratio: float, or int
        The ratio for Lowe's test.

    Returns
    ----------
    static_pts_a: np.array
        An array of x and y coordinates of each self key point on a static object
        which can be matched with acceding key points.

    static_pts_b: np.array
        An array of x and y coordinates of each acceding key point on a static object
        which can be matched with self key points.
   """
    train_idx_dict = {}
    query_idx_dict = {}
    for matches in raw_matches:
        if len(matches) == 2 and matches[0].distance < matches[1].distance * ratio:
            if train_idx_dict.get(matches[0].trainIdx) is None:
                train_idx_dict[matches[0].trainIdx] = []
            train_idx_dict[matches[0].trainIdx].append(matches[0].queryIdx)
            if query_idx_dict.get(matches[0].queryIdx) is None:
                query_idx_dict[matches[0].queryIdx] = []
            query_idx_dict[matches[0].queryIdx].append(matches[0].trainIdx)
    matches = filter_corresponding_points(train_idx_dict, query_idx_dict)
    return matches


def filter_corresponding_points(train_idx_dict, query_idx_dict):
    """ If a point has two matchers then delete this point from consideration.

    Parameters
    ----------
    train_idx_dict: dict
        key - matching points number from the self frame,
        value - matching points number from the acceding frame
            example: {1: [10], 5: [13, 11], ...}
            It means that the self key point №1 was matched with the acceding key point №10.
            The №5 was matched with 2 points, it means that it will be removed.

    query_idx_dict: float, or int
        key - matching points number from the acceding frame,
        value - matching points number from the self frame
            example: {10: [1], 13: [5, 7], ...}
            It means that the acceding key point №10 was matched with the self key point №1.
            The №13 was matched with 2 points, it means that it will be removed.

    Returns
    ----------
    matches: list
        A list of matching points without duplication.

    """
    k_for_del = []
    matches = []
    for key, value in train_idx_dict.items():
        if len(value) > 1:
            k_for_del.append(key)
    for key, value in query_idx_dict.items():
        if len(value) > 1:
            for i in value:
                k_for_del.append(i)
    for i in set(k_for_del):
        del train_idx_dict[i]
    for train_idx, query_idx in train_idx_dict.items():
        matches.append((train_idx, query_idx[0]))
    return matches



"""
EvenVizion library.
https://github.com/AIHunters/EvenVizion

Supporting paper at:
https://github.com/AIHunters/EvenVizion/blob/master/EvenVizion-video_based_camera_localization_component.pdf
This is licensed under an MIT license. See the LICENSE.md file
for more information.

All visualization function using to get stabilization view.
"""
from copy import deepcopy

import cv2
import imutils
import numpy as np

from evenvizion.processing.utils import homography_transformation


def decrease_brightness(background_result, background):
    """ To decrease brightness of previous frame.

        Parameters
        ----------
        background_result: np.array
            An image that is being drawn.

        background: np.array
            The original coordinates of a point.
                format: {"x1": float, "y1": float}

       """
    gray_background = cv2.cvtColor(background_result, cv2.COLOR_BGR2HSV)
    black_background = cv2.cvtColor(background, cv2.COLOR_BGR2HSV)
    v_value = np.array(deepcopy(black_background[..., 2]))
    black_background[..., 2] = np.where((v_value - 2) >= 254, 0, v_value - 2)
    for i, _ in enumerate(gray_background):
        for j in range(0, len(gray_background[i])):
            if gray_background[i][j][2] >= 2:
                gray_background[i][j][2] -= 2
            else:
                gray_background[i][j] = [222, 12.35, 31.76]
    return gray_background, black_background


def change_frame_location(black_background, min_x, x_offset, min_y, y_offset, image_shape):
    """ To draw white border for a video frame.

    Parameters
    ----------
    black_background: np.array
        An image that is being drawn.

    min_x: float
        x coordinate of the upper left corner.

    min_y: float
        y coordinate of the upper left corner.

    x_offset: float
        displacement x coordinate of the upper left corner.

    y_offset: float
        displacement y coordinate of the upper left corner.

    image_shape:
        An original frame shape, at the first place - height, the second - width

    Returns
    ----------
    np.array
        An image with white border for video frame.
    """
    color_background = deepcopy(black_background)
    start_point = (np.abs(min_x) + x_offset, np.abs(min_y) + y_offset)
    end_point = (np.abs(min_x) + x_offset, np.abs(min_y) + y_offset + image_shape[0])
    color_background = cv2.line(color_background, start_point,
                                end_point, color=[255, 255, 255], thickness=1)

    start_point = (np.abs(min_x) + x_offset, np.abs(min_y) + y_offset)
    end_point = (np.abs(min_x) + x_offset + image_shape[1], np.abs(min_y) + y_offset)
    color_background = cv2.line(color_background, start_point,
                                end_point, color=[255, 255, 255], thickness=1)

    start_point = (np.abs(min_x) + x_offset + image_shape[1], np.abs(min_y) + y_offset)
    end_point = (np.abs(min_x) + x_offset + image_shape[1],
                 np.abs(min_y) + y_offset + image_shape[0])
    color_background = cv2.line(color_background, start_point,
                                end_point, color=[255, 255, 255], thickness=1)

    start_point = (np.abs(min_x) + x_offset + image_shape[1],
                   np.abs(min_y) + y_offset + image_shape[0])
    end_point = (np.abs(min_x) + x_offset, np.abs(min_y) + y_offset + image_shape[0])
    color_background = cv2.line(color_background, start_point,
                                end_point, color=[255, 255, 255], thickness=1)
    return color_background


def get_reference_system(homography_dict):
    """ Return dictionary with min and max coordinate in a fixed coordinate system.

    Parameters
    ----------
    homography_dict: dict
        A dictionary with matrix superposition for each frame.
        If you want to analyze displacement only from the previous frame,
        you can specify homography dict without superposition.
            format: {frame_no: [],
                          ...
                      frame_no: []}
    Returns
    ----------
    dict
        The dictionary with min and max coordinate in a fixed coordinate system.
    """
    corner_1 = [0, 0, 1]
    x_corner = []
    y_corner = []
    for _, H in homography_dict.items():
        x_corner.append(homography_transformation(corner_1, H)[0])
        y_corner.append(homography_transformation(corner_1, H)[1])
    corner_dict = {"max_x": int(np.max(x_corner)), "min_x": int(np.min(x_corner)),
                   "max_y": int(np.max(y_corner)),
                   "min_y": int(np.min(y_corner))}
    return corner_dict


def stabilize_view(original_frame, matrix_H, background, corner_dict, width):
    """ To draw a frame position in a fixed coordinate system.

    Parameters
    ----------
    original_frame: np.array
        The frame which position we want to find.

    matrix_H: np.array
        Homography matrix describing the movement of this frame.

    background: np.array
        Background of a fixed coordinate system plane.

    corner_dict: dict
        The dictionary with min and max coordinate in a fixed coordinate system.

    width: int
        Frame width on which homography matrix was received.

    Returns
    ----------
    stabilize_frame: np.array
        The image with frame position on a fixed coordinate system plane.

    background: np.array
        Background of a fixed coordinate system plane with the original frame.
    """
    original_frame = imutils.resize(original_frame, width=width)
    color_shape = [original_frame.shape[0], original_frame.shape[1]]
    transform_vector = (np.dot(matrix_H, [0, 0, 1]))
    x_offset = int(transform_vector[0] / transform_vector[2])
    y_offset = int(transform_vector[1] / transform_vector[2])
    background[np.abs(
        corner_dict["min_y"]) + y_offset:np.abs(corner_dict["min_y"]) + y_offset + color_shape[0],
               np.abs(corner_dict["min_x"]) + x_offset:np.abs(
                   corner_dict["min_x"]) + x_offset + color_shape[1]] = original_frame
    stabilize_frame = change_frame_location(background, corner_dict["min_x"],
                                            x_offset, corner_dict["min_y"], y_offset, color_shape)
    hsv, hsv_background = decrease_brightness(stabilize_frame,
                                              background)
    stabilize_frame = cv2.cvtColor(hsv, cv2.COLOR_HSV2BGR)
    background = cv2.cvtColor(hsv_background, cv2.COLOR_HSV2BGR)
    return stabilize_frame, background


def create_border(image_b):
    """ To create border for a fixed coordinate system plane

    Parameters
    ----------
    image_b: np.array
        Image that is being drawn.

    Returns
    ----------
    np.array:
        The image with border.
    """
    pt1 = (0, 0)
    pt2 = (image_b.shape[1] - 1, image_b.shape[0] - 1)
    image_b = cv2.rectangle(image_b, pt1, pt2, color=[0, 0, 248], thickness=1, lineType=8, shift=0)
    return image_b


def append_text_to_image(image, text, height=300):
    """To put text on a frame

    Parameters
    ----------
    image: np.array
        Image that is being drawn.

    text: str
        The text you want to write on the frame.

    height: int, optional
        If you want a resized image specify height.

    Returns
    ----------
    np.array:
        The image with text.
    """
    image = imutils.resize(image, height=height)
    image = cv2.putText(image, text, (10, 20),
                        fontFace=cv2.FONT_HERSHEY_PLAIN, color=(255, 255, 255),
                        fontScale=1, thickness=1)
    return image


def initialize_background(original_frame, width, corner_dict):
    """To initialize image with fixed coordinate system plane.

    Parameters
    ----------
    corner_dict: dict
         The dictionary with min and max coordinate in fixed coordinate system.

    original_frame: np.array
        The frame from video.

    width: int
        Frame width on which homography matrix was received.

    Returns
    ----------
    np.array:
        The image with a  fixed coordinate system plane.
    """

    image_template = imutils.resize(original_frame, width=width)
    panorama_shape = [int(np.abs(corner_dict["min_y"]) +
                          corner_dict["max_y"] + image_template.shape[0] + 10),
                      int(np.abs(corner_dict["min_x"]) +
                          corner_dict["max_x"] + image_template.shape[1] + 10)]
    background = np.full((panorama_shape[0], panorama_shape[1], 3), 0, dtype=np.uint8)
    background[np.abs(corner_dict["min_y"]):np.abs(corner_dict["min_y"]) + image_template.shape[0],
               np.abs(corner_dict["min_x"]):np.abs(corner_dict["min_x"]) +
               image_template.shape[1]] = image_template
    return background


def create_video_comparison(path_to_video, save_folder, homography_dict, width):
    """To save comparison of a real frame from video with stabilization frame.

    Parameters
    ----------
    path_to_video: str
        The path to the analyzed video.

    save_folder: str
        The path to save pictures with heatmap visualization.

    homography_dict: dict
        Dict with matrix superposition for each frame.
        If you want to analyze displacement analyze only from the previous frame,
        you can specify homography dict without superposition.
            format: {frame_no: [],
                          ...
                      frame_no: []}

    width: int
        Frame width on which homography matrix was received.
    """
    cap = cv2.VideoCapture(path_to_video)
    success, original_frame = cap.read()
    i = 1
    corner_dict = get_reference_system(homography_dict)
    background = initialize_background(original_frame, width, corner_dict)
    while success:
        image_a = append_text_to_image(original_frame, "Original")
        stabilize_frame, background = stabilize_view(original_frame, homography_dict[i], background,
                                                     corner_dict, width)
        image_b = append_text_to_image(stabilize_frame, "EvenVizion")
        image_b = create_border(image_b)
        result = np.concatenate((image_a, image_b), axis=1)
        cv2.imwrite("{}/{}.png".format(save_folder, "%06d" % i), result)
        i += 1
        success, original_frame = cap.read()
        if not success:
            break



"""
EvenVizion library.
https://github.com/AIHunters/EvenVizion

Supporting paper at:
https://github.com/AIHunters/EvenVizion/blob/master/EvenVizion-video_based_camera_localization_component.pdf
This is licensed under an MIT license. See the LICENSE.md file
for more information.

All the functions used to recalculate coordinates into different coordinate system are located here.
"""

from copy import deepcopy
import numpy as np

from evenvizion.processing.utils import inverse_homography_transformation, homography_transformation


def from_original_to_fix(original_coordinates, homography_dict,
                         original_image_shape, resize_image_shape):
    """ To recalculate original coordinates into a fixed coordinate system.

    Parameters
    ----------
    original_coordinates: dict
        The original coordinates of points.
            example: {frame_no: [{"x1": float, "y1": float}, ..., {"x1": float, "y1": float}],
                     ...,
                     frame_no: [{"x1": float, "y1": float}, ..., {"x1": float, "y1": float}]}

    homography_dict: dict
        A dictionary with homography matrices for each frame of video.
            format: {frame_no: {"H": [[.,.,.],[.,.,.],[.,.,.]]}
                    ...
                    frame_no: {"H": [[.,.,.],[.,.,.],[.,.,.]]
                    resize_info: {"h": int, "w": int}}

    original_image_shape: list, or np.array
        An original frame shape, at the first place - height, the second - width.
            example: [224,400].

    resize_image_shape: list, or np.array
        A shape of the frame on which homography matrix was received.
        At the first place - height, the second - width.
            example: [224,400].

    Returns
    ----------
    fixed_coordinates: dict
        A dictionary with original_coordinates in a fixed coordinate system.
            format: {frame_no: [{"x1": float, "y1": float}, ..., {"x1": float, "y1": float}],
                    ...,
                    frame_no: [{"x1": float, "y1": float}, ..., {"x1": float, "y1": float}]}

    """
    original_h, original_w = original_image_shape
    resize_h, resize_w = resize_image_shape
    fixed_coordinates = {}
    h_resize_coefficient = int(resize_h) / original_h
    w_resize_coefficient = int(resize_w) / original_w
    for frame_no, frame_info in original_coordinates.items():
        fixed_coordinates[frame_no] = []
        for rect in frame_info:
            new_rect = deepcopy(rect)
            [new_rect["x1"], new_rect["y1"]] = np.around(homography_transformation(
                [w_resize_coefficient * rect["x1"], h_resize_coefficient * rect["y1"]],
                homography_dict[frame_no]), decimals=2)
            fixed_coordinates[frame_no].append(new_rect)
    return fixed_coordinates


def from_fix_to_original(fix_coordinates, homography_dict,
                         original_image_shape, resize_image_shape):
    """ To recalculate fixed coordinates into a coordinate system relative to the frame.

    Parameters
    ----------
    fix_coordinates: dict
        The original coordinates of points in a fixed system coordinate.
            format: {frame_no: [{"x1": float, "y1": float}, ..., {"x1": float, "y1": float}],
                    ...,
                    frame_no: [{"x1": float, "y1": float}, ..., {"x1": float, "y1": float}]}

    homography_dict: dict
       A dictionary with homography matrices for each frame of video.
            format: {frame_no: {"H": [[.,.,.],[.,.,.],[.,.,.]]}
                    ...
                    frame_no: {"H": [[.,.,.],[.,.,.],[.,.,.]]
                    resize_info: {"h": int, "w": int}}

    original_image_shape: list, or np.array
        An original frame shape, at the first place - height, the second - width.
            example: [224,400].

    resize_image_shape: list, or np.array
        A shape of frame on which homography matrix was received.
        At the first place - height, the second - width.
            example: [224,400].

    Returns
    ----------
    original_coordinates: dict
        A dictionary with original_coordinates in a fixed system coordinate.
            format: {frame_no: [{"x1": float, "y1": float}, ..., {"x1": float, "y1": float}],
                    ...,
                    frame_no: [{"x1": float, "y1": float}, ..., {"x1": float, "y1": float}]}
    """

    original_h, original_w = original_image_shape
    resize_h, resize_w = resize_image_shape
    original_coordinates = {}
    h_resize_coefficient = original_h / resize_h
    w_resize_coefficient = original_w / resize_w
    for frame_no, frame_info in fix_coordinates.items():
        original_coordinates[frame_no] = []
        for rect in frame_info:
            new_rect = deepcopy(rect)
            [new_rect["x1"], new_rect["y1"]] = np.around(inverse_homography_transformation(
                [w_resize_coefficient * rect["x1"], h_resize_coefficient * rect["y1"]],
                homography_dict[frame_no]), decimals=2)
            original_coordinates[frame_no].append(new_rect)
    return original_coordinates



"""
EvenVizion library.
https://github.com/AIHunters/EvenVizion

Supporting paper at:
https://github.com/AIHunters/EvenVizion/blob/master/EvenVizion-video_based_camera_localization_component.pdf
This is licensed under an MIT license. See the LICENSE.md file
for more information.

All the constants used in EvenVizion package.
"""

#: If x or y coordinate is more than this INFINITY_COORDINATE_FLAG,
#: the value of the coordinates is considered undefined.
INFINITY_COORDINATE = 10000

#: If the percentage of the number of points used for
#: calculating homography is less than this value, consider H is None.
LENGTH_ACCOUNTED_POINTS = 0.7

#: The threshold for OpenCV findHomography() function.
THRESHOLD_FOR_FIND_HOMOGRAPHY = 3.0

#: The ratio for Lowe's test.
LOWES_RATIO = 0.5

#: The minimum number of matching points you want to find.
MINIMUM_MATCHING_POINTS = 4

#: The heatmap normalization constant.

HEATMAP_CONSTANT = 1000



"""
EvenVizion library.
https://github.com/AIHunters/EvenVizion

Supporting paper at:
https://github.com/AIHunters/EvenVizion/blob/master/EvenVizion-video_based_camera_localization_component.pdf
This is licensed under an MIT license. See the LICENSE.md file
for more information.

This is an example of the comparison of visualizations between the original video and stabilized video.
"""

import argparse
import os
import cv2
import sys

main_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(main_dir)

main_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.append(main_dir)


from evenvizion.visualization.stabilization import create_video_comparison
from evenvizion.processing.utils import superposition_dict, read_homography_dict

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='custom arguments without using Sacred library')
    parser.add_argument('--path_to_homography_dict', help="path to homography dict",
                        default="test_video_processing/test_video/dict_with_homography_matrix.json")
    parser.add_argument('--path_to_video',
                        default="test_video/test_video.mp4")
    parser.add_argument('--experiment_name', help="folder to save experiment result", default="test_video_processing")
    args = parser.parse_args()
    args.experiment_folder = os.getcwd()
    args.save_folder = args.experiment_folder + "/" + args.experiment_name
    if not os.path.exists(args.save_folder):
        os.makedirs(args.save_folder)
    args.save_folder = args.save_folder + '/{}'.format(os.path.split(args.path_to_video)[-1].split(".")[0])
    if not os.path.exists(args.save_folder):
        os.makedirs(args.save_folder)

    args.save_folder = args.save_folder + "/" + "visualize_camera_stabilization"
    if not os.path.exists(args.save_folder):
        os.makedirs(args.save_folder)

    homography_dict, resize_info = read_homography_dict(args.path_to_homography_dict)
    reformat_homography_dict = superposition_dict(homography_dict)
    cap = cv2.VideoCapture(args.path_to_video)
    create_video_comparison(args.path_to_video, args.save_folder,
                            reformat_homography_dict, resize_info["w"])



"""
EvenVizion library.
https://github.com/AIHunters/EvenVizion

Supporting paper at:
https://github.com/AIHunters/EvenVizion/blob/master/EvenVizion-video_based_camera_localization_component.pdf
This is licensed under an MIT license. See the LICENSE.md file
for more information.

To describe utils functions.
"""

import json

import cv2
import numpy as np
import numpy.linalg as linalg

from evenvizion.processing.constants import INFINITY_COORDINATE, \
    THRESHOLD_FOR_FIND_HOMOGRAPHY, LENGTH_ACCOUNTED_POINTS


class HomographyException(Exception):
    """
        A class used to handle the exception.

        Attributes
        ----------
        message: : str
            Exception description.

        Methods
        ----------
    """

    def __init__(self, message="can't calculate homography matrix"):
        self.message = message
        super().__init__(message)


def remove_double_matching(pts_a, pts_b):
    """ Return only unique key points coordinates.

        Parameters
        ----------
        pts_a: np.array
            An array of x and y coordinates of each key point which matches with points from pts_b.

        pts_b: np.array
            An array of x and y coordinates of each key point which matches with points from pts_a.

        Returns
        ----------
        new_pts_a: np.array
            An array of unique matching points from pts_a.

        new_pts_b: np.array
            An array of unique matching points from pts_b.
        """
    matching_dict = {}
    new_pts_a = []
    new_pts_b = []
    for i, _ in enumerate(pts_a):
        matching_dict[(pts_a[i][0], pts_a[i][1])] = pts_b[i]
    for key, value in matching_dict.items():
        new_pts_a.append(np.array(key))
        new_pts_b.append(value)
    return new_pts_a, new_pts_b


def homography_transformation(vector, matrix_H):
    """ To apply homography transformation with matrix H to vector.

        Parameters
        ----------
         vector: np.array
            An array of x and y vector coordinates, vector[2] should be equal to 1.
                example vector = [21, 457, 1].

        matrix_H: np.array
            Homography matrix.

        Returns
        ----------
        np.array
            An array of x and y vector coordinates after homography transformation.

    """
    while len(vector) < 3:
        vector = np.append(vector, [1])
    new_vector = np.dot(matrix_H, vector)
    return new_vector[:-1] / new_vector[-1]


def inverse_homography_transformation(vector, matrix_H):
    """ To apply inverse homography transformation with matrix H to vector.

        Parameters
        ----------
         vector: np.array
            An array of x and y vector coordinates, vector[2] should be equal to 1.
                example vector = [21, 457, 1].

        matrix_H: np.array
            Homography matrix.

        Returns
        ----------
        np.array
            An array of x and y vector coordinates after inverse homography transformation.
    """
    while len(vector) < 3:
        vector = np.append(vector, [1])
    new_vector = np.dot(linalg.inv(matrix_H), vector)
    return new_vector[:-1] / new_vector[-1]


def matrix_superposition(H, matrix_H_superposition, matrix_H_first=False):
    """ To apply homography transformation with matrix H to vector.

        Parameters
        ----------
        H: np.array
            The first matrix for which we want to get superposition.

        matrix_H_superposition: np.array
            The second matrix for which we want to get superposition.


        matrix_H_first: bool, optional
            if you want to get H as a result then set True,
            if you want to multiply H and matrix_H_superposition then set False.

        Returns
        ----------
        np.array
            Matrix superposition.
    """
    if H is not None:
        if matrix_H_first:
            matrix_H_superposition = H
        else:
            matrix_H_superposition = np.dot(H, matrix_H_superposition)
            matrix_H_superposition = np.divide(matrix_H_superposition, matrix_H_superposition[2][2])
    return matrix_H_superposition


def read_homography_dict(path_to_homography_dict):
    """ To apply homography transformation with matrix H to vector.

        Parameters
        ----------
        path_to_homography_dict: str
            The path to the json with a homography dict.
                json format: {"frame_no": {"H":[]},
                              ...
                              "frame_no": {"H":[]},
                              "resize_info": {"h": (int), "w": (int))}}.
        Returns
        ----------
        homography_dict: dict
            A dict of homography matrix.
                format: {frame_no: {"H":[]},
                              ...
                          frame_no: {"H":[]}}.

        resize_info: dict
            An image shape which is used to get homography dictionary.
                format:{"h": (int), "w": (int))}.
    """
    with open(path_to_homography_dict, "r") as curr_json:
        homography_dict = json.load(curr_json)
        if "resize_info" in homography_dict:
            resize_info = homography_dict["resize_info"]
        else:
            raise ValueError("Specify the height and width of the frame "
                             "for which the homography matrix was obtained")
        resize_info = resize_info
        del homography_dict["resize_info"]
        homography_dict = {int(k): v for k, v in homography_dict.items()}
    return homography_dict, resize_info


def superposition_dict(homography_dict):
    """ To get dict of matrix superposition for each frame.

        Parameters
        ----------
        homography_dict: dict
            A dict of homography matrix.
                format: {frame_no: {"H":[]},
                              ...
                          frame_no: {"H":[]}}.
        Returns
        ----------
        superposition_homography_dict: dict
            A dictionary of matrix superposition for each frame.
                format: {frame_no: [],
                              ...
                          frame_no: []}.
    """

    superposition_homography_dict = {}
    matrix_H_next = None
    H_first = True
    superposition_homography_dict[1] = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    for frame_no, frame_H in homography_dict.items():
        matrix_H_next = matrix_superposition(frame_H["H"], matrix_H_next, H_first)
        H_first = False
        superposition_homography_dict[frame_no] = matrix_H_next
    return superposition_homography_dict


def are_infinity_coordinates(coordinates_value):
    """ To check whether the coordinates are infinite.

        Parameters
        ----------
        coordinates_value: np.array, or list
            x and y coordinates you want to check.

        Returns
        ----------
        bool
            Are coordinates infinite or not.
    """
    try:
        return any(i >= INFINITY_COORDINATE for i in coordinates_value)
    except TypeError:
        return coordinates_value >= INFINITY_COORDINATE


def read_json_with_coordinates(path_to_coordinate):
    """ To read the json file with coordinate.

    Parameters
    ----------
    path_to_coordinate: str
        Path to json with homography dict.
            json format: {"frame_no": [{"x1":x, "y1": y}, ... {"x1":x, "y1": y}],
                          ...
                          "frame_no": [{"x1":x, "y1": y}, ... {"x1":x, "y1": y}]}.

    Returns
    ----------
    coordinates: dict
        A dictionary of coordinates.
            format: {frame_no: [{"x1":x, "y1": y}, ... {"x1":x, "y1": y}],
                     ...
                     frame_no: [{"x1":x, "y1": y}, ... {"x1":x, "y1": y}]}.
    """
    with open(path_to_coordinate, 'r') as path_to_detection:
        coordinates = json.load(path_to_detection)
        coordinates = {int(k): v for k, v in coordinates.items()}
    return coordinates


def get_largest_group_points(r_moving_dict, pts_a, pts_b):
    """ To read the json with coordinate.

    Parameters
    ----------
    r_moving_dict: dict
        A dictionary: key - displacement, value - point number which has this displacement.
            example: {0: [0, 2, 3, ...],
                     ...,
                     10: [1, 5, 7, ...].

    Returns
    ----------
    new_pts_a: np.array
        The coordinates of points which are located in key with the most number of value.
    """

    new_pts_a = []
    new_pts_b = []
    k_max_len = None
    max_len = 0
    for key, value in r_moving_dict.items():
        if len(value) > max_len:
            max_len = len(value)
            k_max_len = key
    for i in r_moving_dict[k_max_len]:
        new_pts_a.append(pts_a[i])
        new_pts_b.append(pts_b[i])
    return np.array(new_pts_a), np.array(new_pts_b)


def find_point_displacement(matrix_H, pts_a, pts_b):
    """ To group points according to their displacement.

    Parameters
    ----------
    matrix_H: np.array
        Matrix describing the change between pts_a and pts_b coordinates.

    pts_a: np.array
        An array of x and y coordinates of each key point which matches with points from pts_b.

    pts_b: np.array
        An array of x and y coordinates of each key point which matches with points from pts_a.


    Returns
    ----------
    r_moving_dict: dict
        A dictionary: key - displacement, value - points number which has this displacement.
            example: {0: [0, 2, 3, ...],
                     ...,
                     10: [1, 5, 7, ...].
    """
    r_moving = []
    r_moving_dict = {}
    if len(pts_a) != len(pts_b):
        raise ValueError("in find_static_part, len(pts_a) != len(pts_b)")
    for i, _ in enumerate(pts_a):
        new_vector = (pts_a[i][0], pts_a[i][1], 1)
        transform_vector = (np.dot(matrix_H, new_vector))
        r_distance = round(np.sum(np.subtract(transform_vector[:2]
                                              / transform_vector[2], pts_b[i]) ** 2) ** 0.5)
        r_moving.append(r_distance)
        if r_moving_dict.get(r_distance) is None:
            r_moving_dict[r_distance] = []
        r_moving_dict[r_distance].append(i)
    return r_moving_dict


def compute_homography(pts_a, pts_b, matrix_H_prev=None):
    """ To compute matrix described the change between static_pts_a and static_pts_b coordinates.

    Parameters
    ----------

    pts_a: np.array
        An array of x and y coordinates of each key point which matches with points from pts_b.

    pts_b: np.array
        An array of x and y coordinates of each key point which matches with points from pts_a.

    matrix_H_prev: np.array, optimal
        If you want to get a matrix describing the change between coordinates in the origin plane,
        you should specify homography matrix superposition
        between origin and frame_a (pts_a located on frame_a).


    Returns
    ----------
    matrix_H_new:  np.array
        Matrix describing the change between static_pts_a and static_pts_b coordinates.
    """
    if matrix_H_prev is not None:
        pts_a = np.apply_along_axis(
            homography_transformation, 1, pts_a, matrix_H_prev)
        pts_b = np.apply_along_axis(
            homography_transformation, 1, pts_b, matrix_H_prev)
    (matrix_H_new, status_new) = cv2.findHomography(
        np.array(pts_a), np.array(pts_b),
        cv2.RANSAC, THRESHOLD_FOR_FIND_HOMOGRAPHY)
    if np.sum(status_new) < LENGTH_ACCOUNTED_POINTS * len(status_new):
        raise HomographyException("not enough points in homography calculation")
    if matrix_H_new is None:
        raise HomographyException()
    return matrix_H_new



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="evenvizion",
    version="0.9.1",
    author="AIHunters",
    author_email="oss@aihunters.com",
    description="EvenVizion - is a video-based camera localization component",
    url="https://github.com/AIHunters/EvenVizion",
    long_description=long_description,
    long_description_content_type="text/markdown",	
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)




"""
EvenVizion library.
https://github.com/AIHunters/EvenVizion

Supporting paper at:
https://github.com/AIHunters/EvenVizion/blob/master/EvenVizion-video_based_camera_localization_component.pdf
This is licensed under an MIT license. See the LICENSE.md file
for more information.

All the visualization functions used during finding a fixed coordinate system.
"""
import os

import cv2
import imutils
import numpy as np

from evenvizion.processing.constants import HEATMAP_CONSTANT
from evenvizion.processing.utils import are_infinity_coordinates, homography_transformation


def draw_matches(image_a, image_b, pts_a, pts_b):
    """ To visualize matching points between two images.

    Parameters
    ----------
    image_a: np.array
     The first image.

    image_b: np.array
     The second image.jr

    pts_a: np.array
            An array of x and y coordinates of each key point
            which matches with points from pts_b.

    pts_b: np.array
         An array of x and y coordinates of each key point
         which matches with points from pts_a.

    Returns
    ----------
    np.array
        Matching points visualization.
    """
    (h_a, w_a) = image_a.shape[:2]
    (h_b, w_b) = image_b.shape[:2]
    visualization = np.zeros((max(h_a, h_b), w_a + w_b, 3), dtype="uint8")
    visualization[0:h_a, 0:w_a] = image_a
    visualization[0:h_b, w_a:] = image_b
    counter = 0
    for (a_point, b_point) in zip(pts_a, pts_b):
        counter += 1
        pt_a = (int(a_point[0]), int(a_point[1]))
        pt_b = (int(b_point[0]) + w_a, int(b_point[1]))
        cv2.line(visualization, pt_a, pt_b, (0, 255, 0), 1)
    return visualization


def find_line_position(i, j, number_of_horizontal_lines,
                       number_of_vertical_lines, shape_info, new_coordinates):
    """ We need to divide the image into parts. We'll call one part a box.
    The number of parts is set using
    the number_of_horizontal_lines and number_of_vertical_lines.
    The total number of boxes is number_of_horizontal_lines * number_of_vertical_lines.
    Boxes are numbered vertically (i) and horizontally (j).

    This function returns image coordinates of (i,j) box's center.

    Parameters
    ----------
    i: int
        The vertical number.

    j: int
        The horizontal number.

    number_of_horizontal_lines: int
        The number of horizontal part.

    number_of_vertical_lines: int
        The number of vertical part.

    shape_info: np.array
        Image shape, at the first place - height, the second - width.
            example: [224,400].

    new_coordinates: np.array, or list
         An array of x and y coordinates in a fixed coordinate system.

    Returns
    ----------
    np.array
       Coordinates of box center.
    """
    height, width = shape_info
    center_r_1 = [
        np.array([int((j * (width) + width / 2) / number_of_vertical_lines),
                  int((i * (height) + height / 2) / number_of_horizontal_lines)]),
        np.array([int((j * (width) + width / 2) / number_of_vertical_lines),
                  (int((i * (height) + height / 2) / number_of_horizontal_lines))]) +
        np.sign(new_coordinates[(int((i * (height) + height / 2) / number_of_horizontal_lines))]
                [int((j * (width) + width / 2) / number_of_vertical_lines)][:-1]) * [10, 0]]
    center_r_2 = [
        np.array([int((j * (width) + width / 2) / number_of_vertical_lines),
                  int((i * (height) + height / 2) / number_of_horizontal_lines)]),
        np.array([int((j * (width) + width / 2) / number_of_vertical_lines),
                  (int((i * (height) + height / 2) / number_of_horizontal_lines))]) +
        np.sign(new_coordinates[(int((i * (height) + height / 2) / number_of_horizontal_lines))]
                [int((j * (width) + width / 2) / number_of_vertical_lines)][:-1]) * [0, 10]]
    return center_r_1, center_r_2


def boxes_drawing(image, number_of_horizontal_lines,
                  number_of_vertical_lines, shape_info, i, j, new_coordinates):
    """ We need to divide the image into parts. We'll call one part box.
    The number of parts is set using
    the number_of_horizontal_lines and number_of_vertical_lines.
    The total number of boxes is number_of_horizontal_lines * number_of_vertical_lines.
    Boxes are numbered vertically (i) and horizontally (j).

    This function draws boxes on image.

    Parameters
    ----------
    image: np.array
         Image that is being drawn.

    i: int
        The vertical number.

    j: int
        The horizontal number.

    number_of_horizontal_lines: int
        The number of horizontal part.

    number_of_vertical_lines: int
        The number of vertical part.

    shape_info: np.array
        Image shape, at the first place - height, the second - width.
            example: [224,400].

    new_coordinates: np.array, or list
        An array of x and y coordinates in fixed coordinate system.

   Returns
   ----------
   np.array
      Image with drawn boxes.
   """
    height, width = shape_info
    cv2.rectangle(image,
                  (int(j * width / number_of_vertical_lines),
                   int(i * height / number_of_horizontal_lines)),
                  (int((j + 1) * width / number_of_vertical_lines),
                   int((i + 1) * height / number_of_horizontal_lines)),
                  (0, 0, 0), 1)
    center_r_1, center_r_2 = find_line_position(i, j,
                                                number_of_horizontal_lines,
                                                number_of_vertical_lines,
                                                image.shape[:-1], new_coordinates)
    cv2.arrowedLine(image, (int(center_r_1[0][0]), int(center_r_1[0][1]) - 5),
                    (int(center_r_1[1][0]), int(center_r_1[1][1] - 5)), (0, 0, 0), 1)
    cv2.arrowedLine(image, (int(center_r_2[0][0]), int(center_r_2[0][1]) - 5),
                    (int(center_r_2[1][0]), int(center_r_2[1][1] - 5)), (0, 0, 0), 1)
    return image


def define_center_coordinate(new_coordinates, i, j, shape_info, lines):
    """ We need to divide the image into parts. We'll call one part box.
    The number of parts is set using the number_of_horizontal_lines and number_of_vertical_lines.
    The total number of boxes is number_of_horizontal_lines * number_of_vertical_lines.
    Boxes are numbered vertically (i) and horizontally (j).

    This function returns coordinates of (i,j) box's center in fixed coordinate system.

    Parameters
    ----------
    i: int
        The vertical number.

    j: int
        The horizontal number.

    lines: dict
        Dict with number number_of_horizontal_lines and number_of_vertical_lines.
            example: {"horizontal": 5, "vertical": 4}.

    shape_info: np.array
        Image shape, at the first place - height, the second - width.
            example: [224,400].

    new_coordinates: np.array, or list
        An array of x and y coordinates in fixed coordinate system.

   Returns
   ----------
    np.array
       The coordinates of box center in a fixed coordinate system.
   """
    height, width = shape_info
    if are_infinity_coordinates(new_coordinates[int((i * (height) + height / 2) /
                                                    lines["horizontal"])]
                                [int((j * (width) + width / 2) / lines["vertical"])]):
        x_coordinate = "n/a"
        y_coordinate = "n/a"
    else:
        x_coordinate = round(
            new_coordinates[int((i * (height) + height / 2) / lines["horizontal"])]
            [int((j * (width) + width / 2) / lines["vertical"])][0])
        y_coordinate = round(
            new_coordinates[int((i * (height) + height / 2) / lines["horizontal"])]
            [int((j * (width) + width / 2) / lines["vertical"])][1])
    return x_coordinate, y_coordinate


def text_coordinates_on_frame(image, coordinates_value, i, j, shape_info, lines):
    """ We need to divide the image into parts. We'll call one part box.
    The number of parts is set using the number_of_horizontal_lines and number_of_vertical_lines.
    The total number of boxes is number_of_horizontal_lines * number_of_vertical_lines.
    Boxes are numbered vertically (i) and horizontally (j).

    This function writes fixed coordinates of (i,j) box's center on image.

    Parameters
    ----------
    image: np.array
        Image that is being drawn.

    i: int
        The vertical number.

    j: int
        The horizontal number.

    lines: dict
        Dict with number number_of_horizontal_lines and number_of_vertical_lines.
            example: {"horizontal": 5, "vertical": 4}.

    shape_info: np.array
        Image shape, at the first place - height, the second - width.
            example: [224,400].

    coordinates_value: np.array, or list.
        An array of x and y coordinates of boxes centers in a fixed coordinate system.

   Returns
   ----------
    np.array
        Image with drawn coordinates.
   """
    height, width = shape_info
    cv2.putText(image,
                'x {}'.format(coordinates_value[0]),
                (int((j * (width) + width / 2) / lines["vertical"] - 20),
                 int((i * (height) + height / 2) / lines["horizontal"] + 10)),
                fontFace=cv2.FONT_HERSHEY_PLAIN, color=(0, 0, 0),
                fontScale=1, thickness=1)
    cv2.putText(image,
                'y {}'.format(coordinates_value[1]),
                (int((j * (width) + width / 2) / lines["vertical"] - 10),
                 int((i * (height) + height / 2) / lines["horizontal"] - 7)),
                fontFace=cv2.FONT_HERSHEY_PLAIN, color=(0, 0, 0),
                fontScale=1, thickness=1)
    return image


def part_line(new_coordinates, image, number_of_horizontal_lines=4, number_of_vertical_lines=5):
    """ We need to divide the image into parts. We'll call one part box.
    The number of parts is set using the number_of_horizontal_lines and number_of_vertical_lines.
    The total number of boxes is number_of_horizontal_lines * number_of_vertical_lines.

    This function returns image with boxes and their centers coordinates in s fixed coordinate system.

    Parameters
    ----------
    image: np.array
        Image that is being drawn.

    number_of_horizontal_lines: int, optional
        The number of horizontal part.

    number_of_vertical_lines: int, optional
        The number of vertical part.

    new_coordinates: np.array, or list
        An array of x and y coordinates in fixed coordinate system.

   Returns
   ----------
    np.array
        Image with drawn boxes and their coordinates.
   """
    lines = {"horizontal": number_of_horizontal_lines, "vertical": number_of_vertical_lines, }
    for i in range(0, number_of_horizontal_lines):
        for j in range(0, number_of_vertical_lines):
            boxes_drawing(image, number_of_horizontal_lines,
                          number_of_vertical_lines, image.shape[:-1], i, j,
                          new_coordinates)
            coordinates_value = define_center_coordinate(new_coordinates,
                                                         i, j, image.shape[:-1], lines)
            image = text_coordinates_on_frame(image, coordinates_value,
                                              i, j, image.shape[:-1], lines)
    return image


def heatmap_frame_processing(new_coordinates, original_image, path_to_save, resize_width=400,
                             heatmap_constant=HEATMAP_CONSTANT):
    """ Here you can get heatmap visualization with visualization of a fixed coordinate system (FCS).
        For visualization of FCS, we need to divide the image into parts. We'll call one part box.
        The number of parts is set using the number_of_horizontal_lines and number_of_vertical_lines.
       The total number of boxes is number_of_horizontal_lines * number_of_vertical_lines.

       This function saves heatmap and fixed coordinate system visualizations.

    Parameters
    ----------
    new_coordinates: np.array, or list
        An array of x and y coordinates in fixed coordinate system.

    original_image: np.array
        Image that is being drawn.

    path_to_save: str
        Path to save getting image.

    resize_width: int, optional
        To speed up the performance of the script, pictures will be resized to this width.

    heatmap_constant: int, optional
        Normalization constant for heatmap.

   """
    image = imutils.resize(original_image, width=resize_width)
    heatmap = np.power(np.sum(np.power(new_coordinates, 2), axis=-1), 0.5)
    heatmap /= heatmap_constant
    heatmap = np.uint8(255 * heatmap)
    heatmap = cv2.applyColorMap(heatmap, cv2.COLORMAP_JET)
    hif = .8
    superimposed_img = heatmap * hif + image
    part_line(new_coordinates, superimposed_img)
    cv2.imwrite(path_to_save, superimposed_img)


def make_template(shape):
    """ Return an array of each pixels coordinate.

    Parameters
    ----------
    shape: np.array, or list
        Template shape.

    Returns
    ----------
    np.array
        An array of each pixels coordinate.
   """

    pictures_template = np.array([[[0] * shape[2]] * shape[1]] * shape[0])
    for i in range(shape[0]):
        for j in range(shape[1]):
            pictures_template[i][j] = [j, i, 1]
    return pictures_template


def heatmap_video_processing(dict_with_homography_matrix, capture,
                             save_folder, resize_width=400, heatmap_constant=HEATMAP_CONSTANT):
    """ Here you can get heatmap visualization with
    visualization of a fixed coordinate system (FCS) for the whole video.
    The biggest displacement during the whole video is returned.

    Parameters
    ----------
    dict_with_homography_matrix:
        Dict with matrix superposition for each frame.
        If you want to analyze displacement only from the previous frame,
        you can specify homography dict without superposition.
            format: {frame_no: [],
                          ...
                      frame_no: []}

    capture: cv2.VideoCapture
        Analyzed video.

    save_folder: str
        The path to save pictures with heatmap visualization.

    resize_width: int
        To speed up the performance of the script, pictures will be resized to this width.

    heatmap_constant: int, optional
        Normalization constant for heatmap.

   Returns
   ----------
    float
        The biggest displacement during the whole video.
   """
    max_r = []
    _, original_image = capture.read()
    original_image = imutils.resize(original_image, resize_width)
    picture_template = make_template(original_image.shape)
    if not os.path.exists(save_folder):
        os.mkdir(save_folder)
    for (frame_no, frame_H) in dict_with_homography_matrix.items():
        path_to_save = "{}/img{}.png".format(save_folder, "%06d" % frame_no)
        new_pictures = np.apply_along_axis(
            homography_transformation, 2, picture_template, frame_H)

        heatmap_frame_processing(new_pictures.tolist(), original_image,
                                 path_to_save, heatmap_constant=heatmap_constant)
        success, original_image = capture.read()
        if not success:
            continue
        original_image = imutils.resize(original_image, resize_width)
        max_r.append(np.max(new_pictures))
    return np.max(max_r)


def draw_original_and_fixed_coordinate(image, rect_original,
                                       rect_fixed, h_resize_coefficient, w_resize_coefficient):
    """ To draw original and fixed coordinate of the same point.

    Parameters
    ----------
    image: np.array
        Image that is being drawn.

    rect_original: dict
        The original coordinates of point.
            format: {"x1": float, "y1": float}.

    rect_fixed: dict
        The original coordinates of point.
            format: {"x1": float, "y1": float}.

    w_resize_coefficient: int
        Specify frame width on which homography matrix was received.

    h_resize_coefficient: int
        Specify frame height on which homography matrix was received.

   Returns
   ----------
    np.array
        Image with point and its coordinates in both coordinate system.
   """
    cv2.circle(image, (int(rect_original["x1"]), int(rect_original["y1"])),
               2, color=(255, 255, 255), thickness=5, lineType=8)
    cv2.putText(image,
                '{}{}'.format("original: ",
                              (int(rect_original["x1"] * h_resize_coefficient),
                               int(rect_original["y1"] * w_resize_coefficient))),
                (int(rect_original["x1"]), int(rect_original["y1"]) - 30),
                fontFace=cv2.FONT_HERSHEY_PLAIN, color=(255, 255, 255),
                fontScale=1, thickness=1)
    cv2.putText(image,
                '{}{}'.format("fixed: ", (int(rect_fixed["x1"]), int(rect_fixed["y1"]))),
                (int(rect_original["x1"]), int(rect_original["y1"]) - 10),
                fontFace=cv2.FONT_HERSHEY_PLAIN, color=(255, 255, 255),
                fontScale=1, thickness=1)
    return image


def comparison_original_with_fixed_coordinate_video_processing(capture, original_coordinates,
                                                               recalculated_coordinates,
                                                               path_to_save, resize_info):
    """ To draw original and fixed coordinates of the same points on all frames from video.

    Parameters
    ----------
    capture: cv2.VideoCapture
        The analyzed video.

    original_coordinates: dict
        The original coordinates points.
            example: {frame_no: [{"x1": float, "y1": float}, ..., {"x1": float, "y1": float}],
                     ...,
                     frame_no: [{"x1": float, "y1": float}, ..., {"x1": float, "y1": float}]}.

    recalculated_coordinates: dict
        The fixed coordinates of points.
            example: {frame_no: [{"x1": float, "y1": float}, ..., {"x1": float, "y1": float}],
                     ...,
                     frame_no: [{"x1": float, "y1": float}, ..., {"x1": float, "y1": float}]}.

    path_to_save: str
        Path to save received image.

    resize_info: dict
        An image shape which used for getting homography dictionary.
            format:{"h": (int), "w": (int))}.
   """
    if not os.path.exists(path_to_save):
        os.mkdir(path_to_save)
    success, image = capture.read()
    frame_no = 1
    img_w = int(capture.get(cv2.CAP_PROP_FRAME_WIDTH))
    img_h = int(capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
    h_resize_coefficient = resize_info["h"] / img_h
    w_resize_coefficient = resize_info["w"] / img_w
    while success:
        if not success:
            continue
        for (rect_original, rect_fixed) in \
                zip(original_coordinates[frame_no], recalculated_coordinates[frame_no]):
            image = draw_original_and_fixed_coordinate(image, rect_original,
                                                       rect_fixed, h_resize_coefficient,
                                                       w_resize_coefficient)

        path_to_save_frame = "{}/img{}.png".format(path_to_save, "%06d" % frame_no)
        cv2.imwrite(path_to_save_frame, image)
        success, image = capture.read()
        frame_no += 1



"""
EvenVizion library.
https://github.com/AIHunters/EvenVizion

Supporting paper at:
https://github.com/AIHunters/EvenVizion/blob/master/EvenVizion-video_based_camera_localization_component.pdf
This is licensed under an MIT license. See the LICENSE.md file
for more information.

Function to get homography dict for video.
"""
import logging
import os

import cv2
import imutils

logging.basicConfig(level=logging.DEBUG)

from evenvizion.processing.frame_processing import FrameProcessing
from evenvizion.processing.matching import NoMatchesException
from evenvizion.processing.utils import compute_homography, \
    HomographyException, matrix_superposition
from evenvizion.visualization.processing_visualization import draw_matches


def get_homography_dict(capture, resize_width=400, matching_path=None, none_H_processing=True):
    """ Returns a dictionary of homography matrices for each frame of the video.

    Parameters
    ----------

    capture: cv2.VideoCapture
        The analyzed video.

    resize_width: int, optional
        To speed up the performance of the script, pictures will be resized to this width.

    matching_path: str, optional
        If is not None, create matching visualization in this folder.


    none_H_processing: bool, optional
     There are some cases where Homography matrix can't be calculated,
            so you need to choose which script you want to run in this case.
            If set True H = H in the previous step,
            False - H = [1,0,0][0,1,0][0,0,1], it means there is no transformation on this frame.

    Returns
    ----------

    homography_dict: dict
     format: {frame_no: {"H": [[.,.,.],[.,.,.],[.,.,.]]}
              ...
              frame_no: {"H": [[.,.,.],[.,.,.],[.,.,.]]
              resize_info:{"h": int, "w": int}}.
    """
    success, result_image = capture.read()
    homography_dict = {}
    if not success:
        raise ValueError("Problem with video! Can't read first frame")
    result_image = imutils.resize(result_image, width=resize_width)
    i = 1
    matrix_H_superposition = None
    matrix_H_prev = None
    matrix_H_first = True
    while success:
        i += 1
        frame_processing_a = FrameProcessing(result_image)
        success, image_b = capture.read()
        if not success:
            continue
        image_b = imutils.resize(image_b, width=resize_width)
        frame_processing_b = FrameProcessing(image_b)
        try:
            key_points_a, key_points_b = \
                frame_processing_b.concatenate_all_features_types(frame_processing_a)
            if matching_path:
                matching_visualization = \
                    draw_matches(frame_processing_a.frame, frame_processing_b.frame, key_points_a, key_points_b)
                cv2.imwrite(os.path.join(matching_path, "matching_vis_{}.png".format(i)), matching_visualization)

            matrix_H = compute_homography(key_points_a, key_points_b, matrix_H_superposition)
        except NoMatchesException:
            logging.info(NoMatchesException)
            matrix_H = None
        except HomographyException:
            logging.info(HomographyException)
            matrix_H = None

        result_image = image_b

        # matrix_H is None case processing
        if matrix_H is None:
            if none_H_processing:
                matrix_H = matrix_H_prev
            else:
                homography_dict[i] = {"H": None}
        ###

        homography_dict[i] = {"H": matrix_H.tolist()}
        matrix_H_superposition = \
            matrix_superposition(matrix_H, matrix_H_superposition, matrix_H_first)
        matrix_H_prev = matrix_H
        matrix_H_first = False
    homography_dict["resize_info"] = \
        {"h": result_image.shape[0], "w": result_image.shape[1]}
    return homography_dict
